#define STRICT
#include <windows.h>
#include "winsock.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


#define LISTEN_PORT     2522
int cons=0;
int seq=0;      //sequencer

struct Connection {
        SOCKET s;
        int seq;
        int conidx;
        
        char destination[256];
        long filesize;
        char tmpdatafilename[256];
        HFILE hf;
        long databytes;
        
        enum { recv_dest, recv_filesize, recv_data } state;

        Connection(SOCKET ss)
          : s(ss),
            seq((::seq+=2)-2),
            filesize(0),
            hf(HFILE_ERROR),
            databytes(0),
            state(recv_dest)
          {
                  destination[0]='\0';
                  tmpdatafilename[0]='\0';
          }
};

Connection *con[128];   //connections
HWND hwnd=NULL; //handle of main window

SOCKET ss=INVALID_SOCKET;       //server socket


static void socketMessageBox(HWND hwnd, const char *m, const char *title, UINT style) {
        char msg[256];
        int e=WSAGetLastError();
        wsprintf(msg,"%s\nLast socket error code=%d",m,e);
        MessageBox(hwnd,msg,title,style);
}


static void dereg(Connection *c) {
        for(int i=c->conidx; i<cons-1; cons++) {
                con[i]=con[i+1];
                con[i]->conidx=i;
        }
        cons--;
}

void handleConnection(Connection *c, WORD event, WORD /*error*/) {
        if(event==FD_READ) {
readagain:
                switch(c->state) {
                        case Connection::recv_dest: {
                                char buf[2];
                                int bytes=recv(c->s,buf,1,0);
                                if(bytes<0) {
                                        if(WSAGetLastError()==WSAEWOULDBLOCK)
                                                return;
                                } else if(bytes==0) {
                                } else {
                                        if(buf[0]=='\r') {
                                                //ignore
                                        } else if(buf[0]=='\n') {
                                                c->state=Connection::recv_filesize;
                                        } else if(lstrlen(c->destination)<128) {
                                                buf[1]='\0';
                                                lstrcat(c->destination,buf);
                                        }
                                }
                                break;
                        }
                        case Connection::recv_filesize: {
                                char buf[1];
                                int bytes=recv(c->s,buf,1,0);
                                if(bytes<0) {
                                        if(WSAGetLastError()==WSAEWOULDBLOCK)
                                                return;
                                } else if(bytes==0) {
                                } else {
                                        if(buf[0]=='\r') {
                                                //ignore
                                        } else if(buf[0]=='\n') {
                                                c->state=Connection::recv_data;
                                        } else
                                                c->filesize = c->filesize*10 + (buf[0]-'0');
                                }
                                break;
                        }
                        case Connection::recv_data: {
                                if(c->tmpdatafilename[0]=='\0') {
                                        wsprintf(c->tmpdatafilename,"\\spool\\nlpd\\tmp%04d",c->seq);
                                        c->hf = _lcreat(c->tmpdatafilename,0);
                                        if(c->hf==HFILE_ERROR) {
                                                closesocket(c->s);
                                                dereg(c);
                                                char msg[256];
                                                wsprintf(msg,"Could not create temporary file '%s'",c->tmpdatafilename);
                                                MessageBox(hwnd,msg,"pcpd error",MB_OK|MB_ICONSTOP);
                                                delete c;
                                                return;
                                        }
                                }
                                static char buf[4096];
                                int bytes;
                                bytes = recv(c->s,buf,4096,0);
                                if(bytes<0) {
                                        if(WSAGetLastError()==WSAEWOULDBLOCK)
                                                return;
                                } else if(bytes==0) {
                                } else {
                                        c->databytes += bytes;
                                        UINT b=_lwrite(c->hf,buf,bytes);
                                        if(b<bytes) {
                                                _lclose(c->hf);
                                                remove(c->tmpdatafilename);
                                                closesocket(c->s);
                                                dereg(c);
                                                MessageBox(hwnd,"Could not write to temporary file\nMaybe the disk is full?","pcpd error",MB_OK|MB_ICONSTOP);
                                                delete c;
                                                return;
                                        }
                                }
                                break;
                        }
                }
                //on some occations me may receive a close-notify before the last read-notify
                //Try to avoid it by peeking and looping
                unsigned long nread=0;
                int rc=ioctlsocket(c->s,FIONREAD,&nread);
                if(rc==0 && nread>0) {
                        goto readagain;
                }
        } else if(event==FD_CLOSE) {
                closesocket(c->s);
                c->s=INVALID_SOCKET;
                dereg(c);
                if(c->state==Connection::recv_data) {
                        _lclose(c->hf);
                        c->hf=HFILE_ERROR;
                        if(c->databytes<c->filesize) {
                                remove(c->tmpdatafilename);
                        } else {
                                char datafilename[256];
                                char controlfilename[256];
                                wsprintf(datafilename,"\\spool\\nlpd\\dfa%04d",c->seq);
                                wsprintf(controlfilename,"\\spool\\nlpd\\cfa%04d",c->seq+1);
                                rename(c->tmpdatafilename,datafilename);
                                //create control file
                                HFILE hf=_lcreat(controlfilename,0);
                                _lwrite(hf,"Ndest/",6);
                                _lwrite(hf,c->destination,lstrlen(c->destination));
                                _lwrite(hf,"\n",1);
                                _lclose(hf);
                        }
                }
                delete c;
        }
}


#define WM_SERVER       (WM_USER+1)
#define WM_CONNECTION   (WM_USER+2)

void handleServer(SOCKET ss, WORD event, WORD /*error*/) {
        if(event==FD_ACCEPT) {
                //new connection
                sockaddr client;
                int namelen=sizeof(client);
                SOCKET s_client;
                if((s_client=accept(ss,&client,&namelen))==INVALID_SOCKET) {
                        //nothing
                        return;
                }
                if(cons==128) {
                        closesocket(s_client);
                        MessageBox(hwnd,"Too many simultaneous connections","pcpd error",MB_OK|MB_ICONSTOP);
                        return;
                }
                Connection *c=new Connection(s_client);
                if(!c) {
                        MessageBox(hwnd,"Not enough memory to register new connection","pcpd error",MB_OK|MB_ICONSTOP);
                        return;
                }
                c->conidx=cons;
                con[cons++]=c;
                if(WSAAsyncSelect(s_client,hwnd,WM_CONNECTION,FD_READ|FD_CLOSE)) {
                        socketMessageBox(hwnd,"Could not set event notification mask on socket","pcpd error",MB_OK|MB_ICONSTOP);
                }
        }
        return;
}


LRESULT CALLBACK __export pcpdWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
        switch(msg) {
                case WM_CREATE:
                        return 0;
                case WM_QUERYOPEN: {
                        char msg[256];
                        wsprintf(msg,"pcpd (16-bit) version 1.0\n"
                                     "Files received: %d\n"
                                     "Currently receiving %d files\n",
                                     seq/2-cons,
                                     cons
                                );
                        MessageBox(hwnd, msg, "pcpd status",MB_OK);
                        return FALSE;
                }
                case WM_CLOSE: {
                        PostQuitMessage(0);
                        return 0;
                }
                case WM_SERVER:
                        if(ss==wParam)
                                handleServer(ss, WSAGETSELECTEVENT(lParam),WSAGETSELECTERROR(lParam));
                        return 0;
                case WM_CONNECTION: {
                        SOCKET s=(SOCKET)wParam;
                        for(int i=0; i<cons; i++) {
                                if(con[i]->s==s) {
                                        handleConnection(con[i],WSAGETSELECTEVENT(lParam),WSAGETSELECTERROR(lParam));
                                        return 0;
                                }
                        }
                        return 0;
                }
                default:
                        return DefWindowProc(hwnd,msg,wParam,lParam);
        }
}


static int startServer() {
        ss = socket(PF_INET, SOCK_STREAM, 0);
        if(ss==INVALID_SOCKET) {
                socketMessageBox(hwnd, "pcpd: Could not create socket","pcpd error",MB_OK|MB_ICONSTOP);
                PostMessage(hwnd, WM_QUIT,0,0);
                return -1;
        }

        sockaddr_in myname;
        memset(&myname,0,sizeof(myname));
        myname.sin_family = AF_INET;
        myname.sin_port = htons(LISTEN_PORT);
        if(bind(ss,(sockaddr*)&myname,sizeof(myname))) {
                socketMessageBox(hwnd, "pcpd: Could bind socket","pcpd error",MB_OK|MB_ICONSTOP);
                closesocket(ss);
                PostMessage(hwnd, WM_QUIT,0,0);
                return -1;
        }

        if(listen(ss,10)) {
                socketMessageBox(hwnd, "pcpd: Could not start listening","pcpd error",MB_OK|MB_ICONSTOP);
                closesocket(ss);
                PostMessage(hwnd, WM_QUIT,0,0);
                return 0;
        }

        int rc;
        rc = WSAAsyncSelect(ss, hwnd, WM_SERVER, FD_ACCEPT);
        if(rc!=0) {
                socketMessageBox(hwnd, "Could not set event notification mask", "pcpd error", MB_OK|MB_ICONSTOP);
                closesocket(ss);
                return -1;
        }
        return 0;
}

static void stopServer() {
        closesocket(ss);
        ss=INVALID_SOCKET;
}

static int createWindow(HINSTANCE hinst, HINSTANCE hinstPrev) {
        if(!hinstPrev) {
                WNDCLASS wndclass;
                memset(&wndclass,0,sizeof(wndclass));
                wndclass.style = 0;
                wndclass.lpfnWndProc = pcpdWndProc;
                wndclass.cbClsExtra = 0;
                wndclass.cbWndExtra = 0;
                wndclass.hInstance = hinst;
                wndclass.hIcon = LoadIcon(hinst,MAKEINTRESOURCE(100));
                wndclass.hCursor = NULL;
                wndclass.hbrBackground = NULL;
                wndclass.lpszMenuName = 0;
                wndclass.lpszClassName = "pcpd";
                if(!RegisterClass(&wndclass)) {
                        MessageBox(NULL,"Could not register window class","pcpd error",MB_OK|MB_ICONSTOP);
                        return -1;
                }
        }

        hwnd = CreateWindow("pcpd",
                            "pcpd",
                            WS_BORDER|WS_CAPTION|WS_OVERLAPPED|WS_SYSMENU|WS_MINIMIZE,
                            CW_USEDEFAULT,0,
                            CW_USEDEFAULT,0,
                            GetDesktopWindow(),
                            NULL,
                            hinst,
                            0
                           );
        if(!hwnd) {
                MessageBox(NULL, "Could not create window", "pcpd error", MB_OK|MB_ICONSTOP);
                return -1;
        }

        ShowWindow(hwnd,SW_SHOWMINIMIZED|SW_SHOW);

        return 0;
}

static void messageLoop() {
        MSG msg;
        while(GetMessage(&msg,NULL,0,0)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
        }
}

static void destroyWindow() {
        DestroyWindow(hwnd);
}


static void shutdownConnections() {
        int i;
        for(i=0; i<cons; i++) {
                closesocket(con[i]->s);
        }

        //empty msg queue
        MSG msg;
        while(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
        }
}

static void createMessageQueue() {
        //enlarge message queue size to avoid overflows of socket notifications
        for(int cMsg=100; cMsg>0; cMsg--) {
                if(SetMessageQueue(cMsg))
                        return;
        }
}

int FAR PASCAL WinMain(HINSTANCE hinst, HINSTANCE hinstPrev, LPSTR, int) {
        createMessageQueue();
        WSADATA wsadata;
        if(WSAStartup(0x0101,&wsadata)) {
                MessageBox(NULL, "Could not initialize socket library","pcpd error", MB_OK|MB_ICONSTOP);
                return -1;
        }
        if(createWindow(hinst,hinstPrev)) {
                WSACleanup();
                return 0;
        }
        if(startServer()) {
                WSACleanup();
                destroyWindow();
                return 0;
        }

        messageLoop();

        stopServer();
        shutdownConnections();
        destroyWindow();
        WSACleanup();
        
        return 0;
}

